/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

/* 
 * Values in this enum help us keep track of how this connection is
 * currently bound to the server 
 */


#include "nufw_declarations.h"



/* DATA STRUCTURES */


/* the structure containing the actual DB connections*/
typedef struct _SQLconnection {
#ifdef APACHE2X
#if APR_HAS_THREADS
  apr_thread_mutex_t *mutex;
#endif
#endif
#ifdef DB_TYPE_MYSQL
  MYSQL *sql;
#else
  PGconn *sql;
#endif
  char *host;
  unsigned int port;
  char *user;
  char *database;
  int ssl_enabled;
  char *ssl_key;
  char *ssl_cert;
  char *ssl_ca;
  char *ssl_ca_dir;
  char *ssl_cypher;
  struct _SQLconnection *next;
}SQLconnection;

#ifdef LDAP_UID
typedef struct _LDAPUIDconnection {
#ifdef APACHE2X
  apr_thread_mutex_t *mutex;
#endif
  LDAP *conn;
  char *host;
  unsigned int port;
  char *binddn;
  char *base;
  struct _LDAPUIDconnection *next;
}LDAPUIDconnection;
#endif

#ifdef APACHE2X
enum allowdeny_type {
    T_ENV,
    T_ALL,
    T_IP,
    T_HOST,
    T_FAIL
};  

typedef struct {
    apr_int64_t limited;
    union {
        char *from;
        apr_ipsubnet_t *ip;
    } x;
    enum allowdeny_type type;
} allowdeny;

#endif


/* Classifies possible returns when opening a DB connection */
typedef enum {
        SQL_OPENDB_FAIL = 0,
        SQL_OPENDB_SUCCESS,
        SQL_OPENDB_ALREADY
} nufw_opendb_ret;

/* The actual Nufw auth config struct. */
typedef struct {
  int auth_authoritative;       /* Is this auth method the one and only? */
  int enabled;                  /* Is auth_nufw enabled in this directory? */

  unsigned short int timewindow;               /* Numbers of tenth of second around request arrival to match uid */

  char *sqlhost;                   /* Name of the SQL server (or space separated list) */
  unsigned int sqlport;                     /* Port of the SQL server */
  char *sqldatabase;               /*Name of the database to use*/
  char *sqltable;               /*Name of the table to use*/
  char *sqluser;                   /*User*/
  char *sqlpassword;               /*Password*/
  unsigned short int version;
  int ssl_enabled;
  char *ssl_key;
  char *ssl_cert;
  char *ssl_ca;
  char *ssl_ca_dir;
  char *ssl_cypher;

#ifdef LDAP_UID
  int ldapuid_in_use;
  char *ldapuid_host;
  unsigned int ldapuid_port;
  char *ldapuid_binddn;
  char *ldapuid_password;
  char *ldapuid_base;
#endif
#ifdef LDAP_UID_CACHE
  int ldapcacheuid_in_use;
#endif
#ifdef APACHE2X
  apr_array_header_t *auths;
#endif
} auth_nufw_dir_config_rec;

typedef struct {
  unsigned int max_concurrent_sql_conns;
  unsigned int max_similar_sql_conns;
#ifdef APACHE2X
  apr_thread_mutex_t *sql_lock;
  int tokens_enabled;
#endif
#ifdef LDAP_UID
  unsigned int max_concurrent_ldapuid_conns;
#ifdef APACHE2X
  apr_thread_mutex_t *ldap_lock;
#endif
#endif
#ifdef LDAP_UID_CACHE
  unsigned int max_ldapuid_cached_objects;
#endif
} auth_nufw_config_rec;



//const char * auth_nufw_set_timewindow(cmd_parms *cmd, void *dummy, char *time);
static const char * auth_nufw_set_timewindow(cmd_parms *cmd, 
                                             void *pcr, 
                                             const char *time);
//const char * auth_nufw_set_sqlport(cmd_parms *cmd, void *dummy, char *port);
static const char * auth_nufw_set_sqlport(cmd_parms *cmd, 
                                          void *pcr, 
                                          const char *port);
//void auth_nufw_child_init(server_rec *s, pool *p);

#ifdef APACHE2X
static apr_status_t nufw_close_links(void *data);
static apr_status_t nufw_close_links_child(void *data);
#else
void auth_nufw_child_exit(server_rec *s, pool *p);
#endif
SQLconnection *find_connection(request_rec *r,
                               SQLconnection *sql_list, 
                               char *host,
                               unsigned int port, 
                               char *database, 
                               char *user, 
                               char *pass,
                               int ssl_enabled,
                               char *ssl_key,
                               char *ssl_cert,
                               char *ssl_ca,
                               char *ssl_ca_dir,
                               char *ssl_cipher);
int auth_nufw_identify_user(request_rec *r);
int check_sql_params(char *host, 
                     char *user, 
                     char *database,
                     unsigned int port, 
                     char *pass,
                     int ssl_enabled,
                     char *ssl_key,
                     char *ssl_cert,
                     char *ssl_ca,
                     char *ssl_ca_dir,
                     char *ssl_cipher);
int nufw_strcmp(const char *s1, const char *s2);
int sql_ssl_params_match(int enabled1,
                         char *key1,
                         char *cert1,
                         char *ca1, 
                         char *ca_dir1, 
                         char *cypher1,
                         int enabled2,
                         char *key2,
                         char *cert2,
                         char *ca2,
                         char *ca_dir2,
                         char *cypher2);
