/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

#ifdef LDAP_UID_CACHE

//#define _GNU_SOURCE
//#include <string.h>

#include "mod_auth_nufw_cache_ldapuid.h"



int check_ldap_uid_cache(apr_pool_t *ourpool,
                         luc_cache_list *luc,
                         char **user,
                         unsigned long result,
                         char *host,
                         unsigned port,
                         char *binddn,
                         char *base)
{
  if (luc == NULL){
      return 0;
  }
  luc_cache_list *browser=luc;
  while (browser != NULL)
  {
      if ((browser->host != NULL)&&
          (browser->binddn != NULL)&&
          (browser->base != NULL))
      {
          if ((!strcmp(browser->host,host))&&
              (!strcmp(browser->binddn,binddn))&&
              (!strcmp(browser->base,base))&&
              (browser->port==port))
          {
              //We found the correct entry in linked list.
              //Now, lets look for our uid.
              ldap_uid_cache *lucbrowser = browser->luc;
              while (lucbrowser != NULL)
              {
                  if (lucbrowser->uid == result)
                  {
                      if (lucbrowser->user != NULL)
                      {
                          if (ourpool == NULL) //Should never occur. Ourpool is apache's request pool
                            *user = strndup(lucbrowser->user,strlen(lucbrowser->user));
                          else
                            *user = apr_pstrdup(ourpool,lucbrowser->user);
                          return 0;
                      }else{
                          return LDAPCACHE_NULL_CHAR;
                      }
                  }
                  lucbrowser=lucbrowser->next;
              }
              return 0;
          } 
      }
      browser = browser->next;
  }
  return 0;
}



int add_ldap_uid_cache(luc_cache_list **luc,
                       char *user,
                       unsigned long result,
                       char *host,
                       unsigned port,
                       char *binddn,
                       char *base,
                       int max)
{
  luc_cache_list *browser=*luc;
      while (browser != NULL)
      {
          if ((browser->host != NULL) &&
              (browser->binddn != NULL) &&
              (browser->base != NULL))
          {
              if (!strcmp(browser->host,host) &&
                  !strcmp(browser->base,base) &&
                  !strcmp(browser->binddn,binddn) &&
                  browser->port == port)
              {
                  int count=0;
                  ldap_uid_cache *luc_browser = browser->luc;
                  while (luc_browser != NULL)
                  {
                      if (luc_browser->uid == result)
                          return LDAP_CACHE_UID_ALREADY_IN;
                      luc_browser = luc_browser->next;
                      count++;
                  }
                  if (count>=max)
                  {
                      nufw_log_error(APLOG_MARK,
                                     APLOG_WARNING,
                                     NULL,
                                     "NOT caching LDAPUID:Max number reached.Do you need to increase AuthNufwMaxLdapUidCachedObjects?");
                  }

                  char *new_user;
                  ldap_uid_cache *new;
                  new_user = strndup(user,strlen(user));
                  if (new_user == NULL)
                      return LDAP_CACHE_COULDNT_STRNDUP;
                  new = (ldap_uid_cache *)malloc(sizeof(ldap_uid_cache));
                  new->uid = result;
                  new->user = new_user;
                       new->next = browser->luc;
                       (browser->luc)=new;
                  return 0;
              }
          }
          browser = browser->next;
      }
      
  //Here we create a new struct for this connection, as it does not exist
  luc_cache_list *new_list;
  char *newhost,*newbinddn,*newbase;
  newhost=strndup(host,strlen(host));
  newbinddn=strndup(binddn,strlen(binddn));
  newbase=strndup(base,strlen(base));

  if ((newhost == NULL)||(newbase == NULL) || (newbinddn==NULL))
      return LDAP_CACHE_COULDNT_STRNDUP;
  new_list = (luc_cache_list *)malloc(sizeof(luc_cache_list));
  if (new_list == NULL)
      return LDAP_CACHE_COULDNT_MALLOC;

  new_list->host = newhost;
  new_list->binddn = newbinddn;
  new_list->base = newbase;
  new_list->port = port;
  new_list->luc = NULL;

  {
      char *new_user;
      ldap_uid_cache *new;
      new_user = strndup(user,strlen(user));
      if (new_user == NULL)
          return LDAP_CACHE_COULDNT_STRNDUP;
      new = (ldap_uid_cache *)malloc(sizeof(ldap_uid_cache));
      new->uid = result;
      new->user = new_user;
      new->next = NULL;
      (new_list->luc)=new;

  }

  new_list->next = *luc;
  *luc = new_list ;

  return 0;

}


#endif
