/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

#ifdef LDAP_UID


#include "mod_auth_nufw_ldapuid.h"

#include <string.h>

#ifdef APACHE2X
#include "apache20.h"
#else
#include "apache13.h"
#endif


char *nufw_ldapuid_connect(LDAP **conn, char *host, char *binddn, unsigned int port, char *pass)
{
  int err,version=3;
  *conn=ldap_init(host,port);
  if (*conn == NULL)
  {
      //nufw_log_error(APLOG_MARK,APLOG_CRIT,NULL,"Cannot ldap_init()");
      return "ldap_init() returned a NULL pointer";
  }
  if (ldap_set_option(*conn, LDAP_OPT_PROTOCOL_VERSION,&version) == LDAP_OPT_SUCCESS)
  {
      err = ldap_bind_s(*conn,binddn,pass,LDAP_AUTH_SIMPLE);
      if (err != LDAP_SUCCESS){
          if (err == LDAP_SERVER_DOWN ){
              ldap_unbind(*conn);
              *conn=NULL;
              return "Ldap server seems to be down";
          }
          //nufw_log_error(APLOG_MARK,APLOG_CRIT,NULL,"ldap bind error: %s",ldap_err2string(err));
          return ldap_err2string(err);
      }
  }
  return NULL;  
}

int nufw_closeldapuid(LDAP *conn)
{
  ldap_unbind(conn);
  return 0;
}

int nufw_ldapuid_find_entry(apr_pool_t *ourpool,
                            LDAP *conn,
                            long int result,
                            char *base,
                            int nufw_timeout,
                            char **user)
{
  char filter[LDAP_UID_QUERY_SIZE];
  int err, attrs_array_len;
  struct timeval timeout;
  LDAPMessage *res, *ldapresult;
  char ** attrs_array;
  
  if (snprintf(filter,LDAP_UID_QUERY_SIZE - 1,
        "(uid=%ld)",result) >= (LDAP_UID_QUERY_SIZE - 1))
      return LDAPUID_NO_QUERY_STRING;
  timeout.tv_sec = nufw_timeout;
  timeout.tv_usec = 0;
  err = ldap_search_st(conn,base,LDAP_SCOPE_SUBTREE,filter,NULL,0,
                       &timeout,&res);
  if (err != LDAP_SUCCESS)
  {
  //    ldap_msgfree(res);
      return LDAPUID_REQUEST_FAILED;
  }
  if (ldap_count_entries(conn,res) == 1 )
  {
      ldapresult = ldap_first_entry(conn,res);
      attrs_array = ldap_get_values(conn,ldapresult,"cn");
      attrs_array_len = ldap_count_values(attrs_array);
      if (attrs_array_len == 1)
      {
          //user = (char *)malloc(sizeof(*attrs_array));
          //if (user == NULL)
          //{
           //   nufw_log_error(APLOG_MARK,APLOG_CRIT,NULL,"nufw_ldapuid_find_entry() : couldn't malloc!");
          //    return LDAPUID_NO_MALLOC;
          //}
          if (ourpool == NULL) // Should never occur, as ourpool is Apache's request pool
            *user = strndup(*attrs_array,strlen(*attrs_array));
          else
            *user = apr_pstrdup (ourpool,*attrs_array);
          ldap_value_free(attrs_array);
          ldap_msgfree(res);
          return 0;
      }
          ldap_value_free(attrs_array);
  }

  ldap_msgfree(res);
  return LDAPUID_PB_NB_RESULT;
}
#endif
