/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/

#include "mod_auth_nufw_sql.h"


#define STATE_OPEN 0x1
#define STATE_ESTABLISHED 0x2


#ifdef DB_TYPE_MYSQL

//This prototype is bad and should be changed to something allowing return
//codes, and SQL pointer
const char *nufw_sql_connectdb(MYSQL **sql,
                         char *host,
                         char *user,
                         char *database,
                         unsigned int port,
                         char *pass,
                         int ssl_enabled,
                         char *ssl_key,
                         char *ssl_cert,
                         char *ssl_ca,
                         char *ssl_ca_dir,
                         char *ssl_cypher)
{
  *sql = mysql_init(*sql);
  if (*sql == NULL) 
    return "NULL pointer";

  if (ssl_enabled)
    mysql_ssl_set(*sql,
                  ssl_key,
                  ssl_cert,
                  ssl_ca,
                  ssl_ca_dir,
                  ssl_cypher);
  
  if (! mysql_real_connect(*sql,host,user,pass,database,port,NULL,0))
//    nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"nufw_sql_connectdb #2 : %s",mysql_error(*sql));
  {
      return (mysql_error(*sql));
  }
  return NULL;
}

unsigned int nufw_sql_thread_safe(void)
{
  return mysql_thread_safe();
}

int nufw_closedb(MYSQL *sql)
{
  mysql_close(sql);
  return 0;
}

int nufw_sql_ping(MYSQL *sql)
{
  return mysql_ping(sql);
}

long int nufw_sql_find_entry(struct apr_pool_t *ourpool,
                            unsigned short int protocol,
                            char *table,
                            struct in_addr addr,
                            unsigned short int sin_port,
                            struct timeval *tv,
                            unsigned short int window,
                            char **user,
                            MYSQL *conn)
{
  char addr_str[INET_ADDRSTRLEN];
  char request[512];
  char timecond_req[64]="";
  char mintime[25],maxtime[25];
  int count=0;
  unsigned long int userid;
  __time_t tmp_time_t;
  MYSQL_RES *result;
  MYSQL_ROW row;
  if (inet_ntop(AF_INET,&addr,addr_str,INET_ADDRSTRLEN) == NULL)
    return SQL_CANT_CONVERT_ADDR;
  if (window > 0)
  {
    tmp_time_t=time_calculate(tv,window,1);
    if ((strftime(mintime,24,"%Y%m%d%H%M%S",localtime(&tmp_time_t)) == 0) ||
        mintime == NULL)
      return SQL_MINTIME_CALCULATE_ERROR;
    tmp_time_t=time_calculate(tv,window,0);
    if ((strftime(maxtime,24,"%Y%m%d%H%M%S",localtime(&tmp_time_t)) == 0) ||
        maxtime == NULL)
      return SQL_MAXTIME_CALCULATE_ERROR;
    if (snprintf(timecond_req,63," AND timestamp<=%s AND timestamp>=%s",maxtime,mintime) >= 63)
        return SQL_CANT_BUILD_REQUEST;
  }
  if (protocol == 1)
  {
    if (snprintf(request,511,"SELECT DISTINCT user_id FROM %s WHERE (tcp_sport=%u AND ip_saddr=%lu AND (state=%d OR state=%d)%s)",
              table,
              htons(sin_port),
              (long unsigned int)htonl(addr.s_addr),
              STATE_OPEN,
              STATE_ESTABLISHED,
              timecond_req)
              >=511)
      return SQL_CANT_BUILD_REQUEST;
  }else if (protocol == 2)
  {
    if (snprintf(request,511,"SELECT DISTINCT username FROM %s WHERE (tcp_sport=%u AND ip_saddr=%lu AND (state=%d OR state=%d)%s)",
              table,
              htons(sin_port),
              (long unsigned int)htonl(addr.s_addr),
              STATE_OPEN,
              STATE_ESTABLISHED,
              timecond_req)
              >=511)
      return SQL_CANT_BUILD_REQUEST;
  }else{
      return SQL_UNKNOWN_PROTOCOL;
  }
//  return 12;  //FIXME This should be removed
  if (mysql_real_query(conn, request, strlen(request)) != 0)
      return SQL_QUERY_ERROR;

  result = mysql_store_result(conn);
  if (result == NULL)
      return SQL_RESULT_ERROR;

  count = mysql_num_rows(result);
  if (count < 1)
  {
      mysql_free_result(result);
      return SQL_NO_MATCH;
  }
  if (count > 1)
  {
      mysql_free_result(result);
      return SQL_TOO_MANY_MATCHES;
  }
  row = mysql_fetch_row(result);
  if (protocol == 1)
  {
    sscanf(row[0],"%lu",&userid); 
    mysql_free_result(result);
    return userid;
  }
  //If we reach this, protocol == 2 in any case
  {
//  int tmp_size = strlen(row[0])+1;
    if (ourpool == NULL) // This should never occur, as ourpool is apache request pool
      *user = strndup(row[0],strlen(row[0]));
    else
      *user = apr_pstrdup(ourpool,row[0]);
//  *user = malloc(sizeof(char) * (tmp_size));
    if (*user == NULL)
      return SQL_INSUFFICIENT_MEM;
//  strncpy(*user,row[0],tmp_size);
    mysql_free_result(result);
    return 0;
  }
}


#else
char *nufw_sql_connectdb(PGconn **sql,
                         char *host,
                         char *user,
                         char *database,
                         unsigned int port,
                         char *pass,
                         int ssl_enabled,
                         char *ssl_key,
                         char *ssl_cert,
                         char *ssl_ca,
                         char *ssl_ca_dir,
                         char *ssl_cypher)
{
  char *pgsql_conninfo;
  char server_port[15];
  int pgsql_status;

  /* Guess why so many programmers write stuff for mysql rather than pg? this
   * API is not as convenient as mysql's IMHO*/
  if (snprintf(server_port,14,"%d",port) >= 14){return "snprintf() failed";}
  pgsql_conninfo = (char *)calloc(strlen(user) + strlen(pass) +
              strlen(host) + strlen(server_port) + strlen(database) +
              strlen("hostaddr='' port= dbname='' user='' password='' ") + 1,
              sizeof(char));
  if (pgsql_conninfo == NULL){return "could not calloc()";}
  strncat(pgsql_conninfo,"host='",6);
  strncat(pgsql_conninfo,host,strlen(host));
  strncat(pgsql_conninfo,"' port=",7);
  strncat(pgsql_conninfo,server_port,strlen(server_port));
  strncat(pgsql_conninfo," dbname='",9);
  strncat(pgsql_conninfo,database,strlen(database));
  strncat(pgsql_conninfo,"' user='",8);
  strncat(pgsql_conninfo,user,strlen(user));
  strncat(pgsql_conninfo,"' password='",12);
  strncat(pgsql_conninfo,pass,strlen(pass));
  strncat(pgsql_conninfo,"'",1);

  *sql = PQconnectdb(pgsql_conninfo);
  pgsql_status = PQstatus(*sql);

  free(pgsql_conninfo);
  if(pgsql_status != CONNECTION_OK) {
    PQfinish(*sql);
    return PQerrorMessage(*sql);
  }
  return NULL;
}

unsigned int nufw_sql_thread_safe(void)
{
  return 1;
}

int nufw_closedb(PGconn *sql)
{
  PQfinish(sql);
  return 0;
}

int nufw_sql_ping(PGconn *sql) 
{
  switch(PQstatus(sql))
  {
    case CONNECTION_STARTED: /* This all sucks*/
    case CONNECTION_MADE:
    case CONNECTION_AUTH_OK:
      return 0;
      break;
    default:
      return 1;
  }
}


long int nufw_sql_find_entry(struct apr_pool_t *ourpool,
                            unsigned short int protocol,
                            char *table,
                            struct in_addr addr,
                            unsigned short int sin_port,
                            struct timeval *tv,
                            unsigned short int window,
                            char **user,
                            PGconn *conn)
{
//  char addr_str[INET_ADDRSTRLEN];
  char request[512];
  char timecond_req[64]="";
  char mintime[25],maxtime[25];
  char tmp_inet1[41];
  int count=0;
  char *str_res;
  __time_t tmp_time_t;
  PGresult *result;
//  MYSQL_ROW row;
  unsigned long int res;
  strncpy(tmp_inet1,inet_ntoa(addr),40) ;
  //if (inet_ntop(AF_INET,&addr,addr_str,INET_ADDRSTRLEN) == NULL)
  //  return SQL_CANT_CONVERT_ADDR;
  if (window > 0) // If window is 0 then we totally ignore time related conditions.
    {
    tmp_time_t=time_calculate(tv,window,1);
    if ((strftime(mintime,24,"%Y%m%d%H%M%S",localtime(&tmp_time_t)) == 0) ||
        mintime == NULL)
      return SQL_MINTIME_CALCULATE_ERROR;
    tmp_time_t=time_calculate(tv,window,0);
    if ((strftime(maxtime,24,"%Y%m%d%H%M%S",localtime(&tmp_time_t)) == 0) ||
        maxtime == NULL)
      return SQL_MAXTIME_CALCULATE_ERROR;
    if (snprintf(timecond_req,63," AND timestamp<=%s AND timestamp>=%s",maxtime,mintime) >= 63)
      return SQL_CANT_BUILD_REQUEST;
    }
  if (protocol == 1)
  {
    if (snprintf(request,511,"SELECT DISTINCT user_id FROM %s WHERE (tcp_sport=%u AND ip_saddr='%s' AND (state=%d OR state=%d)%s)",
              table,
              sin_port,
              tmp_inet1,
              STATE_OPEN,
              STATE_ESTABLISHED,
              timecond_req)
              >=511)
      return SQL_CANT_BUILD_REQUEST;
  }else if (protocol == 2)
  {
    if (snprintf(request,511,"SELECT DISTINCT username FROM %s WHERE (tcp_sport=%u AND ip_saddr='%s' AND (state=%d OR state=%d)%s)",
              table,
              sin_port,
              tmp_inet1,
              STATE_OPEN,
              STATE_ESTABLISHED,
              timecond_req)
              >=511)
      return SQL_CANT_BUILD_REQUEST;
  }else{
    return SQL_UNKNOWN_PROTOCOL;
  }
//  return 12;  //FIXME This should be removed
  result = PQexec(conn, request);
  if (PQresultStatus(result) != PGRES_COMMAND_OK)
  {
      PQclear(result);
      return SQL_QUERY_ERROR;
  }
  if (result == NULL)
  {
      PQclear(result);
      return SQL_QUERY_ERROR;
  }

  count = PQntuples(result);
  if (count < 1)
  {
      PQclear(result);
      return SQL_NO_MATCH;
  }
  if (count > 1)
  {
      PQclear(result);
      return SQL_TOO_MANY_MATCHES;
  }

  if (PQgetisnull(result,0,0))
  {
      PQclear(result);
      return SQL_NO_MATCH;
  }

  str_res = PQgetvalue(result,0,0);
  PQclear(result);
  if (protocol == 1)
  {
    res = ntohl(*((uint32_t *) str_res));
    if (res < 0)
      return SQL_NO_MATCH;

    return(res);
  }
  //If this is reached, protocol == 2 in any case
  {
//  int tmp_size = strlen(str_res);
    if (ourpool == NULL) // THis should never occur as ourconn is Apache's request pool
      *user = strndup(str_res,strlen(str_res));
    else
      *user = apr_pstrdup(ourpool,str_res);
//  *user = malloc(sizeof(char) * (tmp_size));
  if (*user == NULL)
      return SQL_INSUFFICIENT_MEM;
//  strncpy(*user,str_res,tmp_size);
  return 0;
  }
}



#endif


__time_t time_calculate(struct timeval *tv, unsigned short int window, int plusmoins)
{
  struct timeval *ourtime = tv;
  if (plusmoins == 0) /*Are we increasing or decrasing time?*/
  {
    if (ourtime->tv_usec<(100000 - window * 50000)) /*Simple check not to get out of bounds.*/
    {
      ourtime->tv_usec += (window * 50000); /* 50000 = 100 000 /2. The window is half before, half after the event*/
    }else{
      ourtime->tv_usec -= (window * 50000);
      ourtime->tv_sec++;
    }
  }else{
    if (ourtime->tv_usec>=(window * 50000)) /*Second bounds check*/
    {
      ourtime->tv_usec -= (window * 50000);
    }else{
      ourtime->tv_usec += (window * 50000);
      ourtime->tv_sec--; /*We are *pretty* sure this should always be >1, right? At least till 2037 ;-) */
    }
  }
  return ourtime->tv_sec;
}


