/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.*/


/* This module (c) Vincent Deffontaines <vincent.deffontaines@inl.fr>, April 2004.
 * This module is designed to work with Nufw, available at www.nufw.org.
 * Provides User identification as part of a real SSO solution.
 * 
 * A lot of inspiration from other modules
 * A lot of advices from #apache-modules 
 * Decembre 2005 : added AuthNufwAuthFrom directive, implemented with code
 * mostly stolen from mod_access*/



#include "mod_auth_nufw.h"
#include "mod_auth_nufw_sql.h"
#ifdef LDAP_UID
#include "mod_auth_nufw_ldapuid.h"
#endif
#ifdef LDAP_UID_CACHE
#include "mod_auth_nufw_cache_ldapuid.h"
#endif

#ifdef APACHE22
module AP_MODULE_DECLARE_DATA mod_authn_nufw;
#else
#ifdef APACHE20
module AP_MODULE_DECLARE_DATA mod_auth_nufw;
#else
module MODULE_VAR_EXPORT mod_auth_nufw;
#endif
#endif
//
// Nasty nasty global variable to access pchild's pool for 1.3
SQLconnection *child_sqlconnections=NULL;
#ifdef LDAP_UID
LDAPUIDconnection *child_ldapuidconnections=NULL;
#endif
#ifdef LDAP_UID_CACHE
luc_cache_list *child_luc_cache_list=NULL;
#endif

//static void *create_nufw_config(pool *p, server_rec *s)
//{
/*    auth_nufw_server_conf *c =
    (auth_nufw_server_conf *) apr_pcalloc(p, sizeof(auth_nufw_server_conf));
    
    c->mtx = ap_create_mutex(NULL);
    c->sqlconnections = NULL;

    return c;*/
//}

int nufw_strcmp(const char *s1, const char *s2) /*This is the same as strcmp() except it returns 
                                      0 instead of crashing if both pointers are NULL*/
                                    /* It returns non-zero if only one pointer
                                     * is NULL*/
{
  if ((s1 == NULL) && (s2 == NULL))
      return 0;
  if (s1 == NULL)
      return -1;
  if (s2 == NULL)
      return 1;
  return strcmp(s1,s2);
}


int check_sql_params(char *host, 
                     char *user, 
                     char *database,
                     unsigned int port, 
                     char *pass,
                     int ssl_enabled,
                     char *ssl_key,
                     char *ssl_cert,
                     char *ssl_ca,
                     char *ssl_ca_dir,
                     char *ssl_cypher)
{
  if ((host == NULL) || strlen(host)<1)
      return SQL_BAD_HOST;
  if ((user == NULL) || strlen(user)<1)
      return SQL_BAD_USER;
  if ((database == NULL) || strlen(database)<1)
      return SQL_BAD_DATABASE;
  if ((port <1) || (port>65535 ))
      return SQL_BAD_PORT;
  if ((pass == NULL) || strlen(pass)<1)
      return SQL_BAD_PASS;
  if (ssl_enabled)
  {
      if ((ssl_key == NULL) || strlen(ssl_key)<1)
          return SQL_BAD_SSL_KEY;
      if (! access(ssl_key,R_OK))
          return SQL_UNREADABLE_SSL_KEY;
      if ((ssl_cert == NULL) || strlen(ssl_cert)<1)
          return SQL_BAD_SSL_CERT;
      if (! access(ssl_cert,R_OK))
          return SQL_UNREADABLE_SSL_CERT;
      if ((ssl_cypher == NULL) || strlen(ssl_cypher)<1)
          return SQL_BAD_SSL_CYPHER;
      if (ssl_ca != NULL)
          if (! access(ssl_ca,R_OK))
              return SQL_UNREADABLE_CA;
      if (ssl_ca_dir != NULL)
          if (! access(ssl_ca_dir,R_OK&&X_OK))
              return SQL_UNREADABLE_CA_DIR;
  }
  return 0;
}


int sql_ssl_params_match(int enabled1,
                         char *key1,
                         char *cert1,
                         char *ca1, 
                         char *ca_dir1, 
                         char *cypher1,
                         int enabled2,
                         char *key2,
                         char *cert2,
                         char *ca2,
                         char *ca_dir2,
                         char *cypher2) //This is just a basic comparison function. Returns 1 if all match, 0 else*/
{
  if ((!enabled1) && (!enabled2)) //All disabled. Others parameters dont matter, so we consider all match
    return 1;
  if (enabled1 != enabled2)
    return 0;
  if (nufw_strcmp(key1,key2) != 0)
    return 0;
  if (nufw_strcmp(cert1,cert2) != 0)
    return 0;
  if (nufw_strcmp(ca1,ca2) != 0)
    return 0;
  if (nufw_strcmp(ca_dir1,ca_dir2) != 0)
    return 0;
  if (nufw_strcmp(cypher1,cypher2) != 0)
    return 0;
  return 1;
}



int check_ldap_params(char *host,char *binddn,unsigned int port,char *pass)
{
  if ((host == NULL) || strlen(host)<1)
      return LDAP_BAD_HOST;
  if ((binddn == NULL) || strlen(binddn)<1)
      return LDAP_BAD_BINDDN;
  if (port <1)
      return LDAP_BAD_PORT;
  if ((pass == NULL) || strlen(pass)<1)
      return SQL_BAD_PASS;
  return 0;
}


static void *auth_nufw_create_config(apr_pool_t *p, server_rec *s)
{
  auth_nufw_config_rec *c;
  c=(auth_nufw_config_rec *)apr_pcalloc(p, sizeof (auth_nufw_config_rec));
  if (c== (auth_nufw_config_rec *) NULL)
      return NULL;
  c->max_concurrent_sql_conns=DEFAULT_MAX_CONCURRENT_SQL_CONNECTIONS;
  c->max_similar_sql_conns=DEFAULT_MAX_SIMILAR_SQL_CONNECTIONS;
#ifdef APACHE2X
  c->tokens_enabled = 1;
#if APR_HAS_THREADS
  if (! APR_SUCCESS == apr_thread_mutex_create(&c->sql_lock,
                                               APR_THREAD_MUTEX_DEFAULT,
                                               p))
    nufw_log_error(APLOG_MARK,APLOG_WARNING,s,"Could not create SQL list mutex");
#endif
#endif
#ifdef LDAP_UID
  c->max_concurrent_ldapuid_conns=DEFAULT_MAX_CONCURRENT_LDAPUID_CONNECTIONS;
#ifdef APACHE2X
#if APR_HAS_THREADS
  if (! APR_SUCCESS == apr_thread_mutex_create(&c->ldap_lock,
                                               APR_THREAD_MUTEX_DEFAULT,
                                               p))
    nufw_log_error(APLOG_MARK,APLOG_WARNING,s,"Could not create LDAP list mutex");
#endif
#endif
#endif
#ifdef LDAP_UID_CACHE
  c->max_ldapuid_cached_objects=DEFAULT_MAX_LDAPUID_CACHED_OBJECTS;
#endif
  return c;
}


static void *auth_nufw_dir_create_config(apr_pool_t *p, char *d)
{
  auth_nufw_dir_config_rec *sec;
  nufw_log_error(APLOG_MARK,APLOG_DEBUG,NULL,"Entering auth_nufw_dir_create_config");
  sec=(auth_nufw_dir_config_rec *)apr_pcalloc(p, sizeof (auth_nufw_dir_config_rec));
  if (sec== (auth_nufw_dir_config_rec *) NULL)
      return NULL;

  sec->auth_authoritative = 1; /*"Secure" by default*/
  sec->enabled = 0;

  sec->timewindow = DEFAULT_TIME_WINDOW;

  sec->sqlhost = NULL;
#ifdef DB_TYPE_MYSQL
  sec->sqlport = 3306;
#else
  sec->sqlport = 5432;
#endif
  sec->sqldatabase = NULL;
  sec->sqltable = "ulog";
  sec->sqluser = NULL;
  sec->sqlpassword = NULL;
  sec->version = DEFAULT_NUFW_PROTOCOL_VERSION;
  sec->ssl_enabled = 0;
  sec->ssl_key = DEFAULT_SQL_SSL_KEY;
  sec->ssl_cert = DEFAULT_SQL_SSL_CERT;
  sec->ssl_ca = NULL;
  sec->ssl_ca_dir = NULL;
  sec->ssl_cypher = DEFAULT_SQL_SSL_CYPHER;
#ifdef LDAP_UID
  sec->ldapuid_in_use = 0;
  sec->ldapuid_host = NULL;
  sec->ldapuid_port = 389;
  sec->ldapuid_binddn = NULL;
  sec->ldapuid_password = NULL;
  sec->ldapuid_base = NULL;
#endif
#ifdef LDAP_UID_CACHE
  sec->ldapcacheuid_in_use = 0;
#endif

#ifdef APACHE2X
  sec->auths = apr_array_make(p, 1, sizeof(allowdeny));
#endif
  return sec;
}


static const char *
auth_nufw_set_max_sql_conns(cmd_parms *cmd, void *dummy, const char *maxconns)
{
  auth_nufw_config_rec *c =
#ifdef APACHE22
    ap_get_module_config(cmd->server->module_config, &mod_authn_nufw);
#else
    ap_get_module_config(cmd->server->module_config, &mod_auth_nufw);
#endif
  c->max_concurrent_sql_conns = atoi(maxconns);
  if (c->max_concurrent_sql_conns<0) //Set to default value if given is stupid
  {
#ifdef APACHE22
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw : specified AuthnNufwMaxSqlConns is out of acceptable range.");
#else
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwMaxSqlConns is out of acceptable range.");
#endif
      c->max_concurrent_sql_conns = DEFAULT_MAX_CONCURRENT_SQL_CONNECTIONS;
  }
  return NULL;
}

static const char *
auth_nufw_set_max_sim_sql_conns(cmd_parms *cmd, void *dummy, const char *maxconns)
{
  auth_nufw_config_rec *c =
#ifdef APACHE22
    ap_get_module_config(cmd->server->module_config, &mod_authn_nufw);
#else
    ap_get_module_config(cmd->server->module_config, &mod_auth_nufw);
#endif
  c->max_similar_sql_conns = atoi(maxconns);
  if (c->max_concurrent_sql_conns<0) //Set to default value if given is stupid
  {
#ifdef APACHE22
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw : specified AuthnNufwMaxSimilarSqlConns is out of acceptable range.");
#else
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwMaxSimilarSqlConns is out of acceptable range.");
#endif
      c->max_similar_sql_conns = DEFAULT_MAX_SIMILAR_SQL_CONNECTIONS;
  }
  return NULL;
}

#ifdef LDAP_UID
static const char *
auth_nufw_set_max_ldapuid_conns(cmd_parms *cmd, void *dummy, char *maxconns)
{
  auth_nufw_config_rec *c =
#ifdef APACHE22
    ap_get_module_config(cmd->server->module_config, &mod_authn_nufw);
#else
    ap_get_module_config(cmd->server->module_config, &mod_auth_nufw);
#endif
  c->max_concurrent_ldapuid_conns = atoi(maxconns);
  if (c->max_concurrent_ldapuid_conns<1) //Set to default value if given is stupid
  {
#ifdef APACHE22
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw : specified AuthnNufwMaxLdapUidConns is out of acceptable range.");
#else
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwMaxLdapUidConns is out of acceptable range.");
#endif
      c->max_concurrent_ldapuid_conns = DEFAULT_MAX_CONCURRENT_LDAPUID_CONNECTIONS;
  }
  return NULL;
}
#endif

#ifdef LDAP_UID_CACHE
static const char *
auth_nufw_set_max_ldapuid_cached_objs(cmd_parms *cmd, void *dummy, char *maxobjs)
{
  auth_nufw_config_rec *c =
#ifdef APACHE22
    ap_get_module_config(cmd->server->module_config, &mod_authn_nufw);
#else
    ap_get_module_config(cmd->server->module_config, &mod_auth_nufw);
#endif
  c->max_ldapuid_cached_objects = atoi(maxobjs);
  if (c->max_ldapuid_cached_objects<1) //Set to default value if given is stupid
  {
#ifdef APACHE22
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw : specified AuthnNufwMaxLdapUidCachedObjects is out of acceptable range.");
#else
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwMaxLdapUidCachedObjects is out of acceptable range.");
#endif
      c->max_ldapuid_cached_objects = DEFAULT_MAX_LDAPUID_CACHED_OBJECTS;
  }
  return NULL;
}
#endif

static const char *
auth_nufw_set_tokens(cmd_parms *cmd, void *dummy, int flag)
{
  auth_nufw_config_rec *config =
#ifdef APACHE22
    ap_get_module_config(cmd->server->module_config, &mod_authn_nufw);
#else
    ap_get_module_config(cmd->server->module_config, &mod_auth_nufw);
#endif
  config->tokens_enabled = flag;
  return NULL;
}

static const char *
auth_nufw_set_timewindow(cmd_parms *cmd, void *pcr, const char *time)
{
  ((auth_nufw_dir_config_rec *) pcr)->timewindow = atoi(time);            
//  if (cr->timewindow < 0 )
//  {
//      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwTimeWindow is out of acceptable range.");
//      cr->timewindow = DEFAULT_TIME_WINDOW;
//  }
  return NULL;
}

static const char *
auth_nufw_set_sqlport(cmd_parms *cmd, void *pcr , const char *port)
{
//  auth_nufw_dir_config_rec *c = 
//    (auth_nufw_dir_config_rec *)ap_get_module_config(cmd->server->module_config,
//                                                 &mod_auth_nufw);
  auth_nufw_dir_config_rec *cr = (auth_nufw_dir_config_rec *) pcr;
  cr->sqlport = atoi(port);            
  if ((cr->sqlport <= 0 ) || (cr->sqlport > 65536))
  {
#ifdef APACHE22
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw : specified AuthnNufwSQLPort is out of acceptable range.");
#else
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwSQLPort is out of acceptable range.");
#endif
#ifdef DB_TYPE_MYSQL
      cr->sqlport = 3306;
#else
      cr->sqlport = 5432;
#endif
  }
  return NULL;
}

static const char *
auth_nufw_set_protocol_version(cmd_parms *cmd, void *pcr , const char *version)
{
//  auth_nufw_dir_config_rec *c = 
//    (auth_nufw_dir_config_rec *)ap_get_module_config(cmd->server->module_config,
//                                                 &mod_auth_nufw);
  auth_nufw_dir_config_rec *cr = (auth_nufw_dir_config_rec *) pcr;
  cr->version = atoi(version);            
  if ((cr->version <= 0 ) || (cr->version > 2))
  {
#ifdef APACHE22
    nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw : specified AuthnNufwProtocolVersion is out of acceptable range.");
#else
    nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwProtocolVersion is out of acceptable range.");
#endif
    cr->version = DEFAULT_NUFW_PROTOCOL_VERSION;
  }
  return NULL;
}

#ifdef LDAP_UID
static const char *
auth_nufw_set_ldapuid_port(cmd_parms *cmd, auth_nufw_dir_config_rec *cr , char *port)
{
//  auth_nufw_dir_config_rec *c = 
//    (auth_nufw_dir_config_rec *)ap_get_module_config(cmd->server->module_config,
//                                                 &mod_auth_nufw);
  cr->ldapuid_port = atoi(port);            
  if ((cr->ldapuid_port <= 0 ) || (cr->ldapuid_port > 65535))
  {
#ifdef APACHE22
    nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw : specified AuthnNufwLdapUidPort is out of acceptable range.");
#else
    nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw : specified AuthNufwLdapUidPort is out of acceptable range.");
#endif
    cr->ldapuid_port = 389;
  }
  return NULL;
}
#endif

#ifdef APACHE2X
static const char *auth_cmd(cmd_parms *cmd, void *dv, const char *from,
                             const char *where_c)
{
    auth_nufw_dir_config_rec *d = (auth_nufw_dir_config_rec *)dv;
    allowdeny *a;
    char *where = apr_pstrdup(cmd->pool, where_c);
    char *s;
    char msgbuf[120];
    apr_status_t rv;

    if (strcasecmp(from, "from"))
#ifdef APACHE22
        return "AuthnNufwAuthFrom must be followed by 'from'";
#else
        return "AuthNufwAuthFrom must be followed by 'from'";
#endif
        
    a = (allowdeny *) apr_array_push(d->auths);
    a->x.from = where;
    a->limited = cmd->limited;
    
    if (!strncasecmp(where, "env=", 4)) {
        a->type = T_ENV;
        a->x.from += 4;

    }
    else if (!strcasecmp(where, "all")) {
        a->type = T_ALL;
    }
    else if ((s = strchr(where, '/'))) {
        *s++ = '\0';
        rv = apr_ipsubnet_create(&a->x.ip, where, s, cmd->pool);
        if(APR_STATUS_IS_EINVAL(rv)) {
            /* looked nothing like an IP address */
            return "An IP address was expected";
        }
        else if (rv != APR_SUCCESS) {
            apr_strerror(rv, msgbuf, sizeof msgbuf);
            return apr_pstrdup(cmd->pool, msgbuf);
        }
        a->type = T_IP;
    }           
    else if (!APR_STATUS_IS_EINVAL(rv = apr_ipsubnet_create(&a->x.ip, where, NULL, cmd->pool))) {
        if (rv != APR_SUCCESS) {
            apr_strerror(rv, msgbuf, sizeof msgbuf);
            return apr_pstrdup(cmd->pool, msgbuf);
        }   
        a->type = T_IP;
    }   
    else { /* no slash, didn't look like an IP address => must be a host */
        a->type = T_HOST;
    }       
            
    return NULL;
}
#endif


static const command_rec auth_nufw_cmds[] = {
#ifdef LDAP_UID_CACHE
    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwMaxLdapUidCachedObjects",
#else
        "AuthNufwMaxLdapUidCachedObjects", 
#endif
        auth_nufw_set_max_ldapuid_cached_objs,
        NULL, RSRC_CONF, 
        "Maximum number of UIDs to cache per LDAP connection, per child, for uid resolution"),

    AP_INIT_FLAG(
#ifdef APACHE22
        "AuthnNufwLdapUidCacheEnabled", 
#else
        "AuthNufwLdapUidCacheEnabled", 
#endif
        ap_set_flag_slot, 
        (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ldapcacheuid_in_use), 
        OR_AUTHCFG,
        "Set to Off to disable ldap cache in resolution of UIDs grabbed from SQL"),
#endif

#ifdef LDAP_UID
    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwMaxLdapUidConns", 
#else
        "AuthNufwMaxLdapUidConns", 
#endif
        auth_nufw_set_max_ldapuid_conns,
        NULL, RSRC_CONF, 
        "Maximum number of concurrent LDAP connections with given parameters, per child, for uid resolution"),

    AP_INIT_FLAG(
#ifdef APACHE22
        "AuthnNufwLdapUidEnabled", 
#else
        "AuthNufwLdapUidEnabled", 
#endif
        ap_set_flag_slot, 
        (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ldapuid_in_use), 
        OR_AUTHCFG,
        "Set to Off to disable ldap resolution of UIDs grabbed from SQL"
        "Set to On to enable resolutions, and set AuthNufwLdapUid* parameters too"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwLdapUidHost", 
#else
        "AuthNufwLdapUidHost", 
#endif
        ap_set_string_slot,
        (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ldapuid_host), OR_AUTHCFG,
        "Set to the LDAP Host where your UID resolutions will be performed"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwLdapUidPort", 
#else
        "AuthNufwLdapUidPort", 
#endif
        auth_nufw_set_ldapuid_port, NULL, OR_AUTHCFG,
        "Set to the TCP port the LDAP UID resolution server"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwLdapUidBinddn", 
#else
        "AuthNufwLdapUidBinddn", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ldapuid_binddn), OR_AUTHCFG,
         "Set to the DN to bind as on the LDAP UID resolver host"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwLdapUidPassword", 
#else
        "AuthNufwLdapUidPassword", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ldapuid_password), OR_AUTHCFG,
         "Set to the password to use when binding to the Ldap UID resolver host"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwLdapUidBase", 
#else
        "AuthNufwLdapUidBase", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ldapuid_base), OR_AUTHCFG,
         "Set to the base DN on the LDAP UID resolver host"),
#endif

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwMaxSqlConns", 
#else
        "AuthNufwMaxSqlConns", 
#endif
        auth_nufw_set_max_sql_conns,
        NULL, RSRC_CONF, 
        "Maximum number of concurrent SQL connections, per child. 0 means infinite."),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwMaxSimilarSqlConns", 
#else
        "AuthNufwMaxSimilarSqlConns", 
#endif
        auth_nufw_set_max_sim_sql_conns,
        NULL, RSRC_CONF, 
        "Maximum number of _similar_ SQL connections with given parameters, per child. 0 means infinite."),

    AP_INIT_FLAG(
#ifdef APACHE22
        "AuthnNufwAuthoritative", 
#else
        "AuthNufwAuthoritative", 
#endif
        ap_set_flag_slot, 
        (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, auth_authoritative), OR_AUTHCFG,
        "Set to Off to allow access control to be passed along to lower modules if "
        "the UserID cannot be defined by this module"),

#ifdef APACHE2X
    AP_INIT_FLAG(
#ifdef APACHE22
        "AuthnNufwTokensEnabled",
#else
        "AuthNufwTokensEnabled", 
#endif
        auth_nufw_set_tokens,
        NULL, RSRC_CONF, 
        "Whether to mention NuFW in server tokens"),

    AP_INIT_ITERATE2(
#ifdef APACHE22
        "AuthnNufwAuthFrom", 
#else
        "AuthNufwAuthFrom", 
#endif
        auth_cmd, NULL, OR_LIMIT,
        "'from' followed by hostnames or IP-address wildcards"),
#endif

    AP_INIT_FLAG(
#ifdef APACHE22
        "AuthnNufwEnabled", 
#else
        "AuthNufwEnabled",
#endif
        ap_set_flag_slot, 
        (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, enabled), 
        OR_AUTHCFG,
        "Set to Off to disable auth_nufw, even if it's been enabled in a higher tree"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwTimeWindow", 
#else
        "AuthNufwTimeWindow", 
#endif
        auth_nufw_set_timewindow, NULL,
        OR_AUTHCFG,
        "Sets the acceptable time window to identify a connection"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLHost", 
#else
        "AuthNufwSQLHost", 
#endif
        ap_set_string_slot,
        (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, sqlhost), OR_AUTHCFG,
        "Set to the Host where your user tracking SQL base can be reached"),
    
    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwProtocolVersion", 
#else
        "AuthNufwProtocolVersion", 
#endif
        auth_nufw_set_protocol_version, NULL, OR_AUTHCFG,
        "Set to the Nufw protocol version to use"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLPort", 
#else
        "AuthNufwSQLPort", 
#endif
        auth_nufw_set_sqlport, NULL, OR_AUTHCFG,
        "Set to the TCP port the SQL database can be reached"),
    
    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLDatabase", 
#else
        "AuthNufwSQLDatabase", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, sqldatabase), OR_AUTHCFG,
         "Set to the SQL Database to use"),

  AP_INIT_TAKE1(
#ifdef APACHE22
      "AuthnNufwSQLTable", 
#else
      "AuthNufwSQLTable",
#endif
      ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, sqltable), OR_AUTHCFG,
         "Set to the SQL Table to use"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLUser", 
#else
        "AuthNufwSQLUser", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, sqluser), OR_AUTHCFG,
         "Set to the SQL Database username"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLPassword", 
#else
        "AuthNufwSQLPassword", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, sqlpassword), OR_AUTHCFG,
         "Set to the SQL Database password"),

    AP_INIT_FLAG(
#ifdef APACHE22
        "AuthnNufwSQLSSLEnabled", 
#else
        "AuthNufwSQLSSLEnabled", 
#endif
        ap_set_flag_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ssl_enabled), OR_AUTHCFG,
         "Set to On to enable SSL wrapping for SQL connections"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLSSLKeyfile", 
#else
        "AuthNufwSQLSSLKeyfile", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ssl_key), OR_AUTHCFG,
         "Set to the private SSL key for encrypting the SQL connections"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLSSLCertfile", 
#else
        "AuthNufwSQLSSLCertfile", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ssl_cert), OR_AUTHCFG,
         "Set to the public SSL certificate for encrypting the SQL connections"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLSSLCA", 
#else
        "AuthNufwSQLSSLCA", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ssl_ca), OR_AUTHCFG,
         "Set to the file containing your Certificate Authority"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLSSLCAPath", 
#else
        "AuthNufwSQLSSLCAPath", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ssl_ca_dir), OR_AUTHCFG,
         "Set to the directory containing your Certificate Authority files"),

    AP_INIT_TAKE1(
#ifdef APACHE22
        "AuthnNufwSQLSSLCypher", 
#else
        "AuthNufwSQLSSLCypher", 
#endif
        ap_set_string_slot,
         (void *)APR_OFFSETOF(auth_nufw_dir_config_rec, ssl_cypher), OR_AUTHCFG,
         "Set to the cypher you wish to use for encrypting SQL connections"),
    {NULL}
};


#ifdef APACHE2X
void auth_nufw_child_init(apr_pool_t *p, server_rec *s)
#else
void auth_nufw_child_init(server_rec *s, apr_pool_t *p)
#endif
{
#ifdef APACHE2X
  /* Register cleanup hook to close DDB connection (apache 2 doesn't have child_exit) */
  apr_pool_cleanup_register(p, NULL, nufw_close_links, nufw_close_links_child);
#endif
  if (! nufw_sql_thread_safe())
#ifdef APACHE22
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_authn_nufw init : SQL backend is not compiled thread safe");
#else
      nufw_log_error(APLOG_MARK,APLOG_WARNING,NULL,"mod_auth_nufw init : SQL backend is not compiled thread safe");
#endif
}

#ifdef APACHE2X
static int auth_nufw_init_handler(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp,
                                 server_rec *s)
{
  auth_nufw_config_rec *config = 
#ifdef APACHE22
      (auth_nufw_config_rec *)ap_get_module_config(s->module_config,&mod_authn_nufw);
#else
      (auth_nufw_config_rec *)ap_get_module_config(s->module_config,&mod_auth_nufw);
#endif
  if (config->tokens_enabled)
    ap_add_version_component(p, "NuFW");
  return OK;
}
#endif




#ifdef APACHE2X
static apr_status_t nufw_close_links(void *data)
#else
void auth_nufw_child_exit(server_rec *s, apr_pool_t *p)
#endif
{
    /* We must parse the linked list and attempt to close all connections */
    SQLconnection *sc=child_sqlconnections; /* Used to parse the whole linked list */  
    /*sc = ap_get_module_config (r->server->process->pool, &auth_nufw_module);*/
    //sc = ap_get_module_config (p, &auth_nufw_module);
    while (sc != NULL)
    {
        if (nufw_closedb(sc->sql))
        {
#ifndef APACHE2X
            nufw_log_error(APLOG_MARK,APLOG_ERR,s,"couldn't close DB link");
#endif
        }
        sc = sc->next;
    }
#ifdef LDAP_UID
    LDAPUIDconnection *lc=child_ldapuidconnections;
    while (lc != NULL)
    {
        if (nufw_closeldapuid(lc->conn))
        {
#ifndef APACHE2X
            nufw_log_error(APLOG_MARK,APLOG_ERR,s,"couldn't close LDAP UID link");
#endif
        }
        lc = lc->next;
    }
#endif
#ifdef APACHE2X
  return APR_SUCCESS;
#endif
}

#ifdef APACHE2X
static apr_status_t nufw_close_links_child(void *data)
{
  /* do nothing*/
  return APR_SUCCESS;
}
#endif



SQLconnection *find_connection(request_rec *r,
                               SQLconnection *sql_list, 
                               char *host,
                               unsigned int port, 
                               char *database, 
                               char *user, 
                               char *pass,
                               int ssl_enabled,
                               char *ssl_key,
                               char *ssl_cert,
                               char *ssl_ca,
                               char *ssl_ca_dir,
                               char *ssl_cypher)
{
  int test_parms = check_sql_params(host,user,database,port,pass,ssl_enabled,ssl_key,ssl_cert,ssl_ca,ssl_ca_dir,ssl_cypher);
  int conn_nb=0; //Number of counted concurrent connections
  int typed_conn_nb=0; //Number of counted similar connections with the same parameters
  auth_nufw_config_rec *config = 
#ifdef APACHE22
      (auth_nufw_config_rec *)ap_get_module_config(r->server->module_config,&mod_authn_nufw);
#else
      (auth_nufw_config_rec *)ap_get_module_config(r->server->module_config,&mod_auth_nufw);
#endif
  
  if (test_parms != 0)
  {
      char msg[55];
      switch (test_parms) {
        case SQL_BAD_HOST:
          sprintf(msg,"Specified DB host is not valid");
          break;
        case SQL_BAD_USER:
          sprintf(msg,"Specified DB user is not valid");
          break;
        case SQL_BAD_PORT:
          sprintf(msg,"Specified DB port is not valid");
          break;
        case SQL_BAD_DATABASE:
          sprintf(msg,"Specified DB database name is not valid");
          break;
        case SQL_BAD_PASS:
          sprintf(msg,"Specified DB database pass is not valid");
          break;
        case SQL_BAD_SSL_KEY:
          sprintf(msg,"Specified SSL private key is not valid");
          break;
        case SQL_BAD_SSL_CERT:
          sprintf(msg,"Specified SSL public certificate is not valid");
          break;
        case SQL_BAD_SSL_CYPHER:
          sprintf(msg,"Specified SSL cypher is not valid");
          break;
        case SQL_UNREADABLE_CA:
          sprintf(msg,"Specified CA file cannot be read");
          break;
        case SQL_UNREADABLE_CA_DIR:
          sprintf(msg,"Specified CA directory cannot be read or searched");
          break;
        case SQL_UNREADABLE_SSL_KEY:
          sprintf(msg,"Specified SSL private key file not readable");
          break;
        case SQL_UNREADABLE_SSL_CERT:
          sprintf(msg,"Specified SSL public certificate file not readable");
          break;
        default:
          sprintf(msg,"unidentified problem");
          break;
      }
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): Cannot connect to database : %s",msg);
      nufw_log_rerror(APLOG_MARK,APLOG_CRIT,r,"find_connection(): Cannot connect to database : %s at %s:%d",msg,host,port);
      return NULL;
  }


  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Entering find_connection");
  SQLconnection *browser = sql_list;
  SQLconnection *last;
  last=browser;
  if (browser == NULL){
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): browser is null");
  }else{ //browser is not NULL
      if ((browser->host) == NULL)
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): browser->host is null");
      if ((browser->database) == NULL)
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): browser->database is null");
      if ((browser->user) == NULL)
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): browser->user is null");
  }
  while (browser != NULL) 
  {
      if (( browser->host!=NULL) && (browser->database!=NULL) && (browser->user !=NULL))
      {
          nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_conn seeing nb %d : h *%s* p *%d* d *%s* u *%s*",
                                        conn_nb+1,browser->host,browser->port,browser->database,browser->user);
          conn_nb++;
          if ((!(strcmp(host,browser->host) || 
            (port != browser->port) ||
           strcmp(database,browser->database) ||
           strcmp(user,browser->user)
            )) && sql_ssl_params_match(ssl_enabled,
                                       ssl_key,
                                       ssl_cert,
                                       ssl_ca,
                                       ssl_ca_dir,
                                       ssl_cypher,
                                       browser->ssl_enabled,
                                       browser->ssl_key,
                                       browser->ssl_cert,
                                       browser->ssl_ca,
                                       browser->ssl_ca_dir,
                                       browser->ssl_cypher))
          {
#ifdef APACHE2X
#if APR_HAS_THREADS
            typed_conn_nb++;
            nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): now attempting to lock mutex!");
            if (APR_SUCCESS == apr_thread_mutex_trylock(browser->mutex))
            {
              nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): Mutex now locked!");
#endif
#endif
              if (browser->sql == NULL) /* We found a suitable object, but it is dead for some (bad) reason. Try to reconnect*/
              {
                nufw_log_rerror(APLOG_MARK,APLOG_NOTICE,r,"find_connection(): NULL conn found. Shouldn't be. Trying to reinit..");
#ifdef DB_TYPE_MYSQL
                const
#endif
                char *connect_ret=nufw_sql_connectdb(&(browser->sql),
                                                     browser->host,
                                                     browser->user,
                                                     browser->database,
                                                     browser->port,
                                                     pass,
                                                     browser->ssl_enabled,
                                                     browser->ssl_key,
                                                     browser->ssl_cert,
                                                     browser->ssl_ca,
                                                     browser->ssl_ca_dir,
                                                     browser->ssl_cypher);
                if (connect_ret == NULL)
                  return browser;
                nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Failed in reinit..(%s:%d) : %s",browser->host,browser->port,connect_ret);
              }
              /* We found a connection that seems to be suitable. Ping it*/
              if (nufw_sql_ping(browser->sql) == 0)
              {
                /* Database answered! Returning the pointer to our connection*/
                return browser;
              }else{
                /* Database did not answer... Try to reconnect*/
#ifdef DB_TYPE_MYSQL
                const
#endif
                char *connect_ret=nufw_sql_connectdb(&(browser->sql),
                                                   browser->host,
                                                   browser->user,
                                                   browser->database,
                                                   browser->port,
                                                   pass,
                                                   browser->ssl_enabled,
                                                   browser->ssl_key,
                                                   browser->ssl_cert,
                                                   browser->ssl_ca,
                                                   browser->ssl_ca_dir,
                                                   browser->ssl_cypher);
                if (connect_ret == NULL)
                  return browser;
                // No way to reconnect.
                apr_thread_mutex_unlock(browser->mutex); // As an exeption, we unlock the mutex here, as we will return NULL
                nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connecion() unlocked mutex");
                nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Failed in reconnection.. (%s:%d) : %s",browser->host,browser->port,connect_ret);
                last = browser;
                browser = browser->next;
              }
#ifdef APACHE2X
#if APR_HAS_THREADS
            }else{ //Mutex already locked
              last = browser;
              browser = browser->next;
            }
#endif
#endif
          }else{
          /* Conn parameters don't match this conn. Just switch to next connection*/
            last = browser;
            browser = browser->next;
          }
    }else{ //Here browser is not NULL but one of (host,database,user) is NULL. This should never be.
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"find_connection() in trouble. Some parameter is NULL but souldnt. Aborting.");
      last = browser;
      browser = browser->next;
    }
  } // end of while (browser != NULL)
  /* So we found no connection suiting our config. We shall create it now.*/
  if ((conn_nb >= config->max_concurrent_sql_conns) && (config->max_concurrent_sql_conns))
  {
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Not creating a new SQL connection. Already %d of them exist!",host,port,conn_nb);
#ifdef APACHE22
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"   (Maybe you need to increase AuthnNufwMaxSqlConns?)");
#else
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"   (Maybe you need to increase AuthNufwMaxSqlConns?)");
#endif
      return NULL;
  }
  if ((typed_conn_nb >= config->max_similar_sql_conns) && (config->max_similar_sql_conns))
  {
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Not creating a new connection to %s:%d. Already %d of them exist!",host,port,conn_nb);
#ifdef APACHE22
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"   (Maybe you need to increase AuthnNufwMaxSimilarSqlConns?)");
#else
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"   (Maybe you need to increase AuthNufwMaxSimilarSqlConns?)");
#endif
      return NULL;
  }

#ifdef APACHE2X
  browser = (SQLconnection *)apr_palloc(r->server->process->pool,sizeof(SQLconnection));
#else
  browser = (SQLconnection *)malloc(sizeof(SQLconnection));
#endif
#ifdef APACHE2X
#if APR_HAS_THREADS
  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Creating a mutex for brand new conn object.");
  if (! APR_SUCCESS == apr_thread_mutex_create(&browser->mutex,
                          APR_THREAD_MUTEX_DEFAULT,
                          r->server->process->pool))
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Could not create mutex");
  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Creating a mutex for brand new conn object -> done");
#endif
#endif
  browser->sql = NULL;
  browser->host = host;
  browser->port = port;
  browser->database = database;
  browser->user = user;
  browser->ssl_enabled = ssl_enabled;
  browser->ssl_key = ssl_key;
  browser->ssl_cert = ssl_cert;
  browser->ssl_ca = ssl_ca;
  browser->ssl_ca_dir = ssl_ca_dir;
  browser->ssl_cypher = ssl_cypher;

  browser->next = NULL;
    
  if (child_sqlconnections == NULL)
    child_sqlconnections = browser;
  else{
#ifdef APACHE2X
#if APR_HAS_THREADS
  if (APR_SUCCESS == apr_thread_mutex_lock(config->sql_lock))
  {
#endif
#endif
  //Go to the actual end of list... This is a leak to conn_nb as well as
  //typed_conn_nb!
  while(last->next != NULL)
      last = last->next;
  last->next = browser;
#ifdef APACHE2X
#if APR_HAS_THREADS
  apr_thread_mutex_unlock(config->sql_lock);
  }else{
    nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Could not get lock on the SQL conn list! Aborting");
    return NULL;
  }
#endif
#endif
  }
#ifdef APACHE2X
#if APR_HAS_THREADS

  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Trying to get mutex from brand new conn object.");
  if (APR_SUCCESS == apr_thread_mutex_trylock(browser->mutex))
  {
    nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Mutex now locked ! :)");
#endif
#endif

#ifdef DB_TYPE_MYSQL
    const
#endif
    char *connect_ret=nufw_sql_connectdb(&(browser->sql),
                                       browser->host,
                                       browser->user,
                                       browser->database,
                                       browser->port,
                                       pass,
                                       browser->ssl_enabled,
                                       browser->ssl_key,
                                       browser->ssl_cert,
                                       browser->ssl_ca,
                                       browser->ssl_ca_dir,
                                       browser->ssl_cypher);
    if (connect_ret != NULL)
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Failed in connecting to database.. (%s:%d) : %s",browser->host,browser->port,connect_ret);
    return browser;
#ifdef APACHE2X
#if APR_HAS_THREADS
  }else{ // Mutex we just created is already locked!
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Mutex we just created is already locked! Is this a very busy webserver?");
      return NULL;
  }
#endif
#endif
}


#ifdef LDAP_UID
LDAPUIDconnection *find_ldapuidconnection(request_rec *r,LDAPUIDconnection *ldap_list, char *host,unsigned int port, 
                                  char *binddn, char *pass)
{
  int test_parms = check_ldap_params(host,binddn,port,pass);
  int conn_nb=0; //Number of counted concurrent connections of this type
  auth_nufw_config_rec *config = 
#ifdef APACHE22
      (auth_nufw_config_rec *)ap_get_module_config(r->server->module_config,&mod_authn_nufw);
#else
      (auth_nufw_config_rec *)ap_get_module_config(r->server->module_config,&mod_auth_nufw);
#endif
  
  if (test_parms != 0)
  {
      char msg[50];
      switch (test_parms) {
        case LDAP_BAD_HOST:
          sprintf(msg,"Specified LDAP host is not valid");
          break;
        case LDAP_BAD_BINDDN:
          sprintf(msg,"Specified LDAP bind dn is not valid");
          break;
        case LDAP_BAD_PORT:
          sprintf(msg,"Specified LDAP port is not valid");
          break;
        case LDAP_BAD_PASS:
          sprintf(msg,"Specified LDAP password is not valid");
          break;
        default:
          sprintf(msg,"unidentified problem");
          break;
      }
      nufw_log_rerror(APLOG_MARK,APLOG_ERR,r,"find_ldapuidconnection(): Won't connect to ldap tree %s:%d :",host,port);
      nufw_log_rerror(APLOG_MARK,APLOG_ERR,r," %s",msg);
      return NULL;
  }


  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Entering find_ldapuidconnection");
  LDAPUIDconnection *browser = ldap_list;
  LDAPUIDconnection *last;
  last=browser;
  if (browser == NULL){
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_ldapuidconnection() browser is null");
  }else{
      if ((browser->host) == NULL)
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_ldapuidconnection() browser->host is null");
      if ((browser->binddn) == NULL)
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_ldapuidconnection() browser->binddn is null");
  }
  while (browser != NULL) 
  {
      if (( browser->host!=NULL) && (browser->binddn !=NULL))
      {
        nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_ldapuid seeing h *%s* p *%d* u *%s*",
                                        browser->host,browser->port,browser->binddn);
        if (!(strcmp(host,browser->host) || 
            (port != browser->port) ||
           strcmp(binddn,browser->binddn)
            ))
        {
            conn_nb++;
#ifdef APACHE2X
#if APR_HAS_THREADS
            nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): now attempting to lock ldap mutex!");
            if (APR_SUCCESS == apr_thread_mutex_trylock(browser->mutex))
            {
              nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"find_connection(): ldap Mutex now locked!");
#endif
#endif
              if (browser->conn == NULL) /* We found a suitable object, but it is dead for some (bad) reason. Try to reconnect*/
              {
                nufw_log_rerror(APLOG_MARK,APLOG_NOTICE,r,"find_ldapuidconnection(): NULL conn found. Shouldn't be. Trying to reinit..");
                char *conn_ret=nufw_ldapuid_connect(&(browser->conn),
                                                    browser->host,
                                                    browser->binddn,
                                                    browser->port,
                                                    pass);
                if (conn_ret == NULL)
                  return browser;
                apr_thread_mutex_unlock(browser->mutex); // As an exeption, we unlock the mutex here, as we will return NULL
                nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r," Failed in reinit.. (%s:%d) : %s",browser->host,browser->port,conn_ret);
                return NULL;
              }
              return browser;
#ifdef APACHE2X
#if APR_HAS_THREADS
            }else{ //Mutex already locked
              last = browser;
              browser = browser->next;
            }
#endif
#endif
        }else{
          /* Just switch to next connection*/
          last = browser;
          browser = browser->next;
        }
    }else{
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"find_ldapuidconnection() in trouble. Some parameter is NULL but souldnt. Aborting.");
      last = browser;
      browser = browser->next;
    }
  }
  /* So we found no connection suiting our config. We shall create it now.*/
  if (conn_nb >= config->max_concurrent_ldapuid_conns)
  {
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Not creating a new connection to %s:%d. Already %d of them exist!",host,port,conn_nb);
#ifdef APACHE22
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"   (Maybe you need to increase AuthnNufwMaxLdapUidConns?)");
#else
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"   (Maybe you need to increase AuthNufwMaxLdapUidConns?)");
#endif
    return NULL;
  }
#ifdef APACHE2X
  browser = (LDAPUIDconnection *)apr_palloc(r->server->process->pool,sizeof(LDAPUIDconnection));
#else
  browser = (LDAPUIDconnection *)malloc(sizeof(LDAPUIDconnection));
#endif
#ifdef APACHE2X
#if APR_HAS_THREADS
  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Creating a mutex for brand new ldap conn object.");
  if (! APR_SUCCESS == apr_thread_mutex_create(&browser->mutex,
                          APR_THREAD_MUTEX_DEFAULT,
                          r->server->process->pool))
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Could not create ldap mutex");
  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Creating a mutex for brand new ldap conn object -> done");
#endif
#endif
  browser->conn = NULL;
  browser->host = host;
  browser->port = port;
  browser->binddn = binddn;
  browser->next = NULL;
    
  if (child_ldapuidconnections == NULL)
    child_ldapuidconnections = browser;
  else{
#ifdef APACHE2X
#if APR_HAS_THREADS
  if (APR_SUCCESS == apr_thread_mutex_lock(config->ldap_lock))
  {
#endif
#endif
  //Go to the actual end of list... This is a leak to conn_nb on very rare
  //conditions it may lead to few more than max connections up. On the other
  //hand, fixing this "problem" would get to a (much) longer lock, and unoptimized
  //performances.
    while(last->next != NULL)
      last = last->next;
    last->next = browser;
#ifdef APACHE2X
#if APR_HAS_THREADS
  apr_thread_mutex_unlock(config->ldap_lock);
  }else{
    nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Could not get lock on the SQL conn list! Aborting");
    return NULL;
  }
#endif
#endif
  }
#ifdef APACHE2X
#if APR_HAS_THREADS

  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Trying to get mutex from brand new conn object.");
  if (APR_SUCCESS == apr_thread_mutex_trylock(browser->mutex))
  {
    nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Mutex now locked ! :)");
#endif
#endif

  char *conn_ret=nufw_ldapuid_connect(&(browser->conn),
                                                      browser->host,
                                                      browser->binddn,
                                                      browser->port,
                                                      pass);

  if (conn_ret != NULL)
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Failed connecting to LDAP (uid) tree.. (%s:%d) : %s",
                          browser->host,
                          browser->port,
                          conn_ret);
  return browser;
#ifdef APACHE2X
#if APR_HAS_THREADS
  }else{ // Mutex we just created is already locked!
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Mutex we just created is already locked! Is this a very busy webserver?");
      return NULL;
  }
#endif
#endif
}
#endif


#ifdef APACHE2X
static int in_domain(const char *domain, const char *what)
{
    int dl = strlen(domain);
    int wl = strlen(what);

    if ((wl - dl) >= 0) {
        if (strcasecmp(domain, &what[wl - dl]) != 0)
            return 0;
    
        /* Make sure we matched an *entire* subdomain --- if the user
         * said 'allow from good.com', we don't want people from nogood.com
         * to be able to get in.
         */
    
        if (wl == dl)
            return 1;           /* matched whole thing */
        else
            return (domain[0] == '.' || what[wl - dl - 1] == '.');
    }
    else
        return 0;
}


static int authornot(request_rec *r, apr_array_header_t *a, int method)
{
  allowdeny *ap = (allowdeny *) a->elts;
  apr_int64_t mmask = (AP_METHOD_BIT << method);
  int i;
  int gothost = 0;
  const char *remotehost = NULL;
  for (i = 0; i < a->nelts; ++i) {
        if (!(mmask & ap[i].limited))
            continue;
  switch (ap[i].type) {
        case T_ENV:
            if (apr_table_get(r->subprocess_env, ap[i].x.from)) {
                return 1;
            }
            break;

        case T_ALL:
            return 1;

        case T_IP:
            if (apr_ipsubnet_test(ap[i].x.ip, r->connection->remote_addr)) {
                return 1;
            }
            break;

        case T_HOST:
            if (!gothost) {
                int remotehost_is_ip;

                remotehost = ap_get_remote_host(r->connection, r->per_dir_config,
                                                REMOTE_DOUBLE_REV, &remotehost_is_ip);

                if ((remotehost == NULL) || remotehost_is_ip)
                    gothost = 1;
                else
                    gothost = 2;
            }

            if ((gothost == 2) && in_domain(ap[i].x.from, remotehost))
                return 1;
            break;

        case T_FAIL:
            /* do nothing? */
            break;
        }
    }

    return 0;
}
#endif  


int auth_nufw_identify_user(request_rec *r)
{
  /*Where is our configuration? */
  nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Entering auth_nufw_identify_user");
  int method = r->method_number;
  long int result = 0;
#ifdef LDAP_UID_CACHE
  auth_nufw_config_rec *config = 
#ifdef APACHE22
      (auth_nufw_config_rec *)ap_get_module_config(r->server->module_config,&mod_authn_nufw);
#else
      (auth_nufw_config_rec *)ap_get_module_config(r->server->module_config,&mod_auth_nufw);
#endif
  if (config == (auth_nufw_config_rec *)NULL)
  {
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Problem at auth_nufw_identify user init with *config");
    return DECLINED;
  }
#endif
  auth_nufw_dir_config_rec *sec = 
#ifdef APACHE22
    (auth_nufw_dir_config_rec *)ap_get_module_config(r->per_dir_config, &mod_authn_nufw);
#else
    (auth_nufw_dir_config_rec *)ap_get_module_config(r->per_dir_config, &mod_auth_nufw);
#endif
  if (sec == (auth_nufw_dir_config_rec *)NULL)
  {
    nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"Problem at auth_nufw_identify user init with *sec");
    return DECLINED;
  }
  if (!sec->enabled)
    return DECLINED;


#ifdef APACHE2X
  if (!authornot(r, sec->auths, method))
        return DECLINED;

#endif  

  struct timeval *tv=NULL;
  struct timezone *tz=NULL;

  if (sec->timewindow > 0) //No need to initialize these if they are not going to be used
  {
    tv=(struct timeval *)malloc(sizeof(struct timeval));
    tz=(struct timezone *)malloc(sizeof(struct timezone));
    if (gettimeofday(tv,tz) != 0)
    {
      nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"auth_nufw_identify_user() Could not get an answer from gettimeofday().");
      free(tz);
      free(tv);
      return sec->auth_authoritative? HTTP_UNAUTHORIZED : DECLINED;
    }
    free(tz);
  }
  
  /*Where is our (child bound) database links pool ?*/
//  SQLconnection *conns = child_sqlconnections;
//    (struct SQLconnection *)ap_get_module_config(r->server->process->pool, &auth_nufw_module);
  /* First thing we do, find an appropriate database connection*/
  SQLconnection *our_conn = find_connection(r,
                                            child_sqlconnections,
                                            sec->sqlhost,
                                            sec->sqlport,
                                            sec->sqldatabase,
                                            sec->sqluser,
                                            sec->sqlpassword,
                                            sec->ssl_enabled,
                                            sec->ssl_key,
                                            sec->ssl_cert,
                                            sec->ssl_ca,
                                            sec->ssl_ca_dir,
                                            sec->ssl_cypher);
  if ((our_conn == NULL)||(our_conn->sql == NULL)) /* If this test is passed, we are positively sure we have a functionnal db access*/
  {
      //if (sec->timewindow > 0) //What is best for performance? run this test
      //or call free() on a NULL pointer?
#ifdef APACHE2X
#if APR_HAS_THREADS
      if (our_conn != NULL) // Not gonna unlock a mutex that doesn't exist, right?
      {
        apr_thread_mutex_unlock(our_conn->mutex);
        nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"auth_nufw_identify_user() unlocked mutex");
      }
#endif
#endif
        free(tv);
      return sec->auth_authoritative? HTTP_UNAUTHORIZED : DECLINED;
  }


  /* Now, we work on :
   * r->connection->remote_addr which is a sockaddr_in structure.
   * r->connection->remote_addr->sin_port would be the port,
   * r->connection->remote_addr->sin_addr would be the internet address*/
    /*To be secure, let's unset the user field, which should already be anyway*/
#ifdef APACHE2X
    r->user = NULL;
#else
    r->connection->user = NULL;
#endif
    char *user = NULL;
    result =
    nufw_sql_find_entry(r->pool,
                       sec->version,
                       sec->sqltable,
#ifdef APACHE2X
                       r->connection->remote_addr->sa.sin.sin_addr,
                       r->connection->remote_addr->sa.sin.sin_port,
#else
                       r->connection->remote_addr.sin_addr,
                       r->connection->remote_addr.sin_port,
#endif
                       tv,
                       sec->timewindow,
                       &user,
                       our_conn->sql); /* Build SQL request and perform it*/

#ifdef APACHE2X
#if APR_HAS_THREADS
      apr_thread_mutex_unlock(our_conn->mutex);
      nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"auth_nufw_identify_user() unlocked mutex");
#endif
#endif

    //if (sec->timewindow > 0) //What is best for performance? run this test or
    //call free() on a NULL pointer?
      free(tv);
    if ((sec->version == 1) && (result >= 0) ){
#ifdef LDAP_UID
      if (sec->ldapuid_in_use)
      {
#ifdef LDAP_UID_CACHE
        if (sec->ldapcacheuid_in_use){
          nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"LDAP UIDs cache... Gonna try It.");
          check_ldap_uid_cache(r->pool,
                           child_luc_cache_list,
                           &user,
                           result,
                           sec->ldapuid_host,
                           sec->ldapuid_port,
                           sec->ldapuid_binddn,
                           sec->ldapuid_base);
        }
        if (user == NULL) //Get through this whether ldapcacheuid_in_use or not
        {
          nufw_log_rerror(APLOG_MARK,APLOG_DEBUG,r,"Nothing for this in LDAP UIDs cache... Gonna try an LDAP request.");
#endif
          LDAPUIDconnection *our_ldapuid_conn = find_ldapuidconnection(r,
                                                               child_ldapuidconnections,
                                                               sec->ldapuid_host,
                                                               sec->ldapuid_port,
                                                               sec->ldapuid_binddn,
                                                               sec->ldapuid_password);
          if ((our_ldapuid_conn == NULL)||(our_ldapuid_conn->conn == NULL))
          {
#ifdef APACHE2X
#if APR_HAS_THREADS
            if (our_ldapuid_conn != NULL) // Not gonna unlock some mutex if it does not exist
              apr_thread_mutex_unlock(our_ldapuid_conn->mutex);
#endif
#endif
            return sec->auth_authoritative? HTTP_UNAUTHORIZED : DECLINED;
          }
          if (nufw_ldapuid_find_entry(r->pool,
                            our_ldapuid_conn->conn,
                            result,
                            sec->ldapuid_base,
                            LDAP_UID_TIMEOUT,
                            &user) != 0)
          {
#ifdef APACHE2X
#if APR_HAS_THREADS
            apr_thread_mutex_unlock(our_ldapuid_conn->mutex);
#endif
#endif
            nufw_log_rerror(APLOG_MARK,APLOG_NOTICE,r,"auth_nufw_identify_user(): Error from nufw_ldapuid_find_entry()");
            return sec->auth_authoritative? HTTP_UNAUTHORIZED : DECLINED;
          }
#ifdef APACHE2X
#if APR_HAS_THREADS
          apr_thread_mutex_unlock(our_ldapuid_conn->mutex);
#endif
#endif

          if (user == NULL)
          {
            nufw_log_rerror(APLOG_MARK,APLOG_WARNING,r,"No resolution found in LDAP for user! Numeric ID %ld",result);
            return sec->auth_authoritative? HTTP_UNAUTHORIZED : DECLINED;
          }
#ifdef APACHE2X
          r->user = user;
#else
          r->connection->user = user;
#endif
#ifdef LDAP_UID_CACHE
          if (sec->ldapcacheuid_in_use)
            add_ldap_uid_cache(&child_luc_cache_list,
                           user,
                           result,
                           sec->ldapuid_host,
                           sec->ldapuid_port,
                           sec->ldapuid_binddn,
                           sec->ldapuid_base,
                           config->max_ldapuid_cached_objects);
#endif
          return OK;
#ifdef LDAP_UID_CACHE
        } else {
#ifdef APACHE2X
          r->user = user;
#else
          r->connection->user = user;
#endif
          return OK;
        }
#endif
      }else{

#endif
          /* WHAT THE Fsck is this test about? return value is wrong typed btw
          if (result==0){
                return user;
          }*/
        user = (char *) apr_pcalloc(r->pool, 12); /* 12 is ugly but bigger than any longint->char*/
        if (user == NULL)
        {
          nufw_log_rerror(APLOG_MARK,APLOG_ERR,r,"ran out of memory. apr_pcalloc() failed");
          return sec->auth_authoritative? HTTP_UNAUTHORIZED : DECLINED;
        }
        sprintf(user,"%lu",result);
#ifdef APACHE2X
        r->user = user;
#else
        r->connection->user = user;
#endif
        return OK;
#ifdef LDAP_UID
      }
#endif
    }
    if (result < 0)
    {
      char msg[50];
      switch (result) {
        case SQL_CANT_CONVERT_ADDR:
          sprintf(msg,"error: inet_ntop could not convert address");
          break;
        case SQL_MINTIME_CALCULATE_ERROR:
          sprintf(msg,"error calculating min time");
          break;
        case SQL_MAXTIME_CALCULATE_ERROR:
          sprintf(msg,"error calculating max time");
          break;
        case SQL_CANT_BUILD_REQUEST:
          sprintf(msg,"error building SQL request");
          break;
        case SQL_QUERY_ERROR:
          sprintf(msg,"error performing SQL request");
          break;
        case SQL_RESULT_ERROR:
          sprintf(msg,"error in SQL result");
          break;
        case SQL_NO_MATCH:
          sprintf(msg,"no suitable result found in database");
          break;
        case SQL_TOO_MANY_MATCHES:
          sprintf(msg,"too many suitable results found in database");
          break;
        case SQL_UNKNOWN_PROTOCOL:
          sprintf(msg,"weird problem : unknown protocol specified");
          break;
        case SQL_INSUFFICIENT_MEM:
          sprintf(msg,"could not allocate memory");
          break;
        default:
          sprintf(msg,"an unidentified error occured");
          break;
        }
          
        nufw_log_rerror(APLOG_MARK,APLOG_NOTICE,r,"auth_nufw_identify_user() : %s",msg);
        return sec->auth_authoritative? HTTP_UNAUTHORIZED : DECLINED;
    }

    // This code is reached only on sec->version == 2 and result >=0
#ifdef APACHE2X
    r->user = user;
#else
    r->connection->user = user;
#endif
    return OK;
}



#ifdef APACHE2X
static void register_hooks(apr_pool_t *p) {
    ap_hook_child_init(auth_nufw_child_init, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_check_user_id(auth_nufw_identify_user,NULL,NULL,APR_HOOK_FIRST);
    ap_hook_post_config(auth_nufw_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
};

#ifdef APACHE22
module AP_MODULE_DECLARE_DATA mod_authn_nufw = {
#else
module AP_MODULE_DECLARE_DATA mod_auth_nufw = {
#endif
    STANDARD20_MODULE_STUFF,
    auth_nufw_dir_create_config, /* create per-directory config structures */
    NULL,                        /* merge per-directory config structures */
    auth_nufw_create_config,     /* create per-server config structures */
    NULL,                        /* merge per-server config structures*/
    auth_nufw_cmds,              /* command table */
    register_hooks               /* register hooks */
};
#else
module MODULE_VAR_EXPORT mod_auth_nufw =
{
    STANDARD_MODULE_STUFF,
    NULL,                       /* initializer */
    auth_nufw_dir_create_config,  /* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    auth_nufw_create_config,  /* server config */
    NULL,                       /* merge server configs */
    auth_nufw_cmds,               /* command table */
    NULL,                       /* handlers */
    NULL,                       /* filename translation */
    auth_nufw_identify_user,    /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    auth_nufw_child_init,       /* child_init */
    auth_nufw_child_exit,       /* child_exit */
    NULL                        /* post read-request */
};
#endif //#ifdef APACHE2X
